/**************************************************************************************************************************
// This script was authored by Lorne Brooks - Email: lbtools2k@gmail.com
//
// A tool for visualising the projected lower and upper masses (pelvis and base of the neck respectively) in relation to the feet of a
// standard Biped rig. Markers between the feet show where the masses intersect, which may help with determining where the static
// weight is distributed when posing. It does not take into account motion dynamics and so is mainly used for when standing still.
// Additionally, when a foot is raised above the ground it's line markers turn blue. Likewise if lowered below the ground they turn red.
// Otherwise they stay green when level with the ground (+ or - 0.1 unit tollerance).
//
// To add a Biped Mass Marker to a Biped, simply select any single Biped part and click the toolbar icon. After the markers are imported
// into the scene, a selection set is created containing the frozen markers. To delete the markers and selection set, hold SHIFT and click
// the toolbar icon.
//
//
// REVISION HISTORY:
// Version 1.1 21/11/18;
//						- Markers now scale to match the biped size when merged.
//						- Added the ability to delete all markers (if multiple) or by picking a single Biped with markers via Shift+Click on the
//						  toolbar icon.
// Version 1.0 01/04/18;
//						- First public release.
//
//
// This script is not copyrighted, please feel free to modify or distribute.
//
**************************************************************************************************************************/

global g_LBTools_BMM

struct s_LBTools_bipedMassMarkers
(

/*==================================================================================
// Structure variables
====================================================================================*/
	
BMMmaxfile = getDir #plugcfg + @"\LBTools_BipedMassMarkers.max",
	
/*==================================================================================
// Structure functions
/*==================================================================================

f_addConstraints
f_findHierarchyRoot
f_deleteBMMlayer
f_collectBipRoots
f_deleteAllMassMarkers
f_getNode
f_scaleParts
f_bipedMassMarkers
	
====================================================================================*/

fn f_addConstraints sourceObj targetObj relative:false =
(
	sourceObj.pos.controller = Position_List()
	sourceObj.pos.controller.available.controller = Position_Constraint()
	sourceObj.pos.controller[2].appendTarget targetObj 50.0
	sourceObj.pos.controller.active = 2
	
	sourceObj.rotation.controller = Rotation_List()
	sourceObj.rotation.controller.available.controller = Orientation_Constraint()
	sourceObj.rotation.controller[2].appendTarget targetObj 50.0
	sourceObj.rotation.controller.active = 2
	
	if relative do
	(
		sourceObj.pos.controller.Position_Constraint.controller.RELATIVE = on
		sourceObj.rotation.controller.Orientation_Constraint.controller.RELATIVE = on
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_findHierarchyRoot obj =
(
	if obj != undefined do 
	(
		if isValidnode obj.parent then
		(
			f_findHierarchyRoot obj.parent
		)
		else
		(
			obj
		)
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_deleteBMMlayer bipRoot =
(
	local layerName = ("BIPED MASS MARKERS" + " - " + bipRoot.name)
	local layer = layerManager.getLayerFromName layerName
	if layer != undefined do
	(
		layer.nodes &nodesInLayer
		delete nodesInLayer
		layerManager.deleteLayerByName layerName
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_collectBipRoots =
(
	local a_bipRoots = #()
	for b in geometry where (classof b.controller == Vertical_Horizontal_Turn) do
	(
		append a_bipRoots b
	)
	a_bipRoots
),
	
/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_deleteAllMassMarkers =
(
	local a_bipRoots = g_LBTools_BMM.f_collectBipRoots()
	for bipRoot in a_bipRoots do
	(
		g_LBTools_BMM.f_deleteBMMlayer bipRoot
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_getNode a_objs bRoot str =
(
	local index = findItem a_objs (getNodeByName (str + "_" +  bRoot.name))
	a_objs[index]
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_scaleParts a_objs bipRoot =
(
	local STANDARD_HEIGHT = 167.864 --// Biped height 190.0 generic units: head height in figure mode
	
	local bipHead = biped.getNode bipRoot #Head
	
	bipRoot.controller.figureMode = true
	local headHeight = bipHead.transform.position.z
	bipRoot.controller.figureMode = false
	
	local scaleFactor = headHeight / STANDARD_HEIGHT

	--// Resize the main scaler dummy
	local BMM_Scaler = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_Scaler"
	scale BMM_Scaler [scaleFactor,scaleFactor,scaleFactor]
	
	--// Resize the render dots by scaling the xforms for the dot meshes
	setCommandPanelTaskMode #modify
	toolMode.coordsys #world
	
	local BMM_FootCentre_Dot = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_FootCentre_Dot"
	select BMM_FootCentre_Dot 
	subobjectLevel = 1
	modPanel.setCurrentObject BMM_FootCentre_Dot.modifiers[#XForm]
	BMM_FootCentre_Dot.modifiers[#XForm].gizmo.scale = [scaleFactor,scaleFactor,scaleFactor]
		
	local BMM_MassCentre_Dot = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_MassCentre_Dot"
	select BMM_MassCentre_Dot
	subobjectLevel = 1
	modPanel.setCurrentObject BMM_MassCentre_Dot.modifiers[#XForm]
	BMM_MassCentre_Dot.modifiers[#XForm].gizmo.scale = [scaleFactor,scaleFactor,scaleFactor]
	
	clearselection()
	
	--// Resize the render lines
	local BMM_FootToFoot_Line = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_FootToFoot_Line"
	BMM_FootToFoot_Line.render_thickness *= scaleFactor
	
	local BMM_LFoot_MassLine = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_LFoot_MassLine"
	BMM_LFoot_MassLine.render_thickness *= scaleFactor
	
	local BMM_UpperToLower_Line = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_UpperToLower_Line"
	BMM_UpperToLower_Line.render_thickness *= scaleFactor
	
	local BMM_RFoot_MassLine = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_RFoot_MassLine"
	BMM_RFoot_MassLine.render_thickness *= scaleFactor
	
	local BMM_LFoot_Line = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_LFoot_Line"
	BMM_LFoot_Line.render_thickness *= scaleFactor
	
	local BMM_RFoot_Line = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_RFoot_Line"
	BMM_RFoot_Line.render_thickness *= scaleFactor
	
	local BMM_UpperMass_Line = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_UpperMass_Line"
	BMM_UpperMass_Line.render_thickness *= scaleFactor
	
	local BMM_LowerMass_Line = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_LowerMass_Line"
	BMM_LowerMass_Line.render_thickness *= scaleFactor
	
	--// Adjust the helper sizes
	local a_helpers = for h in helpers where (findstring h.name "BMM_") != undefined collect h
	for obj in a_helpers do
	(
		if classOf obj == Dummy do obj.boxsize *= scaleFactor
		if classOf obj == Point OR classOf obj == ExposeTm do obj.size *= scaleFactor
	)
	
	--// Adjust the min and max thresholds for changing line colour when raising and lowering foot markers
	local val = 0.1 * scaleFactor
	local str_formula = "if World_Position_Z > " + (val as string) + " then [0,0,255] else (if World_Position_Z < " + ((-val) as string) + " then [255,0,0] else [0,255,0])"
	local BMM_LFoot_ExpTM = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_LFoot_ExpTM"
	local BMM_RFoot_ExpTM = g_LBTools_BMM.f_getNode a_objs bipRoot "BMM_RFoot_ExpTM"
	local mat = BMM_FootToFoot_Line.material
	paramWire.connect BMM_LFoot_ExpTM.baseObject[#World_Position_Z] mat[2][#Shader_Basic_Parameters][#Diffuse_Color] str_formula
	paramWire.connect BMM_RFoot_ExpTM.baseObject[#World_Position_Z] mat[3][#Shader_Basic_Parameters][#Diffuse_Color] str_formula
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

/************************************************************************************************
// Note:
// For multiple Biped rigs in the scene, each one must have a unique root name for multiple BMMs.
*************************************************************************************************/
fn f_bipedMassMarkers =
(
	if doesFileExist g_LBTools_BMM.BMMmaxfile then
	(
		if selection.count == 1 AND classOf selection[1] == Biped_Object then
		(
			local bipRoot = g_LBTools_BMM.f_findHierarchyRoot selection[1]
			mergeMAXFile g_LBTools_BMM.BMMmaxfile #useMergedMtlDups #neverReparent #select
			
			--// Rename the scene layer to a unique name based on the chosen biped
			local layer = layerManager.getLayerFromName "BIPED MASS MARKERS"
			layer.setName ("BIPED MASS MARKERS" + " - " + bipRoot.name)
			
			-- // Check if the BMM helper already exist in the scene for this Biped
			local moduleExists = false
			for h in helpers do
			(
				if (findString h.name ("BMM_Floor_Pt" + "_" + bipRoot.name)) != undefined do
				(
					moduleExists = true
				)
			)
			
			if moduleExists == false then
			(		
				-- // Rename the BMM module to avoid naming conflicts with multiple characters, for expose TM assignment - based on Biped root name as the suffix
				local a_BMMobjs = for s in selection collect s
				for s in a_BMMobjs do
				(
					s.name = s.name + "_" + bipRoot.name
				)
			
				--// Scale the markers to match the biped size
				g_LBTools_BMM.f_scaleParts a_BMMobjs bipRoot
				
				-- // Assign the renamed BMM objects to common variables
				local BMM_Floor_Pt = undefined
				local BMM_Floor2_Pt = undefined
				local BMM_LFoot_Pt = undefined
				local BMM_RFoot_Pt = undefined
				local BMM_Pelvis_Pt = undefined
				local BMM_Ribcage_Pt = undefined
				local BMM_LFoot_ExpTM = undefined
				local BMM_RFoot_ExpTM = undefined
				local BMM_Pelvis_ExpTM = undefined
				local BMM_Ribcage_ExpTM = undefined
				local BMM_LFoot_Line = undefined
				local BMM_RFoot_Line = undefined
				local BMM_LowerMass_Line = undefined
				local BMM_UpperMass_Line = undefined
				local BMM_UpperToLower_Line = undefined
				local BMM_LFoot_MassLine = undefined
				local BMM_RFoot_MassLine = undefined
				local BMM_FootToFoot_Line = undefined
				local BMM_MassCentre_Dot = undefined
				local BMM_FootCentre_Dot = undefined
				
				for s in a_BMMobjs do
				(
					if (findString s.name "BMM_Floor_Pt") != undefined do (BMM_Floor_Pt = s)
					if (findString s.name "BMM_Floor2_Pt") != undefined do (BMM_Floor2_Pt = s)
					if (findString s.name "BMM_LFoot_Pt") != undefined do (BMM_LFoot_Pt = s)
					if (findString s.name "BMM_RFoot_Pt") != undefined do (BMM_RFoot_Pt = s)
					if (findString s.name "BMM_Pelvis_Pt") != undefined do (BMM_Pelvis_Pt = s)
					if (findString s.name "BMM_Ribcage_Pt") != undefined do (BMM_Ribcage_Pt = s)
					if (findString s.name "BMM_LFoot_ExpTM") != undefined do (BMM_LFoot_ExpTM = s)
					if (findString s.name "BMM_RFoot_ExpTM") != undefined do (BMM_RFoot_ExpTM = s)
					if (findString s.name "BMM_Pelvis_ExpTM") != undefined do (BMM_Pelvis_ExpTM = s)
					if (findString s.name "BMM_Ribcage_ExpTM") != undefined do (BMM_Ribcage_ExpTM = s)
					if (findString s.name "BMM_LFoot_Line") != undefined do (BMM_LFoot_Line = s)
					if (findString s.name "BMM_RFoot_Line") != undefined do (BMM_RFoot_Line = s)
					if (findString s.name "BMM_LowerMass_Line") != undefined do (BMM_LowerMass_Line = s)
					if (findString s.name "BMM_UpperMass_Line") != undefined do (BMM_UpperMass_Line = s)
					if (findString s.name "BMM_UpperToLower_Line") != undefined do (BMM_UpperToLower_Line = s)
					if (findString s.name "BMM_LFoot_MassLine") != undefined do (BMM_LFoot_MassLine = s)
					if (findString s.name "BMM_RFoot_MassLine") != undefined do (BMM_RFoot_MassLine = s)
					if (findString s.name "BMM_FootToFoot_Line") != undefined do (BMM_FootToFoot_Line = s)
					if (findString s.name "BMM_MassCentre_Dot") != undefined do (BMM_MassCentre_Dot = s)
					if (findString s.name "BMM_FootCentre_Dot") != undefined do (BMM_FootCentre_Dot = s)
				)
				
				-- // Rename the material based on Biped root name
				BMM_UpperMass_Line.material.name = "BMM - " + bipRoot.name
				
				-- // Find the needed Biped parts
				bPelvis = biped.getNode bipRoot #Pelvis
				bNeck = biped.getNode bipRoot #Neck
				bLFoot = biped.getNode bipRoot #lLeg link:3
				bRFoot = biped.getNode bipRoot #rLeg link:3
				
				-- // Add constraints to BMM objects, binding them to the Biped parts
				g_LBTools_BMM.f_addConstraints BMM_LFoot_Pt bLFoot
				g_LBTools_BMM.f_addConstraints BMM_RFoot_Pt bRFoot
				g_LBTools_BMM.f_addConstraints BMM_Pelvis_Pt bPelvis
				g_LBTools_BMM.f_addConstraints BMM_Ribcage_Pt bNeck
				
				-- // Zero-off the feet helper positions to the floor
				BMM_LFoot_Pt.pos.controller.Available.controller = Position_XYZ()
				BMM_LFoot_Pt.pos.controller.active = 3
				BMM_LFoot_Pt.pos.z = 0.0
				BMM_RFoot_Pt.pos.controller.Available.controller = Position_XYZ()
				BMM_RFoot_Pt.pos.controller.active = 3
				BMM_RFoot_Pt.pos.z = 0.0
					
				-- // Assign Biped objects to Expose Node
				BMM_LFoot_ExpTM.exposeNode = BMM_LFoot_Pt
				BMM_RFoot_ExpTM.exposeNode = BMM_RFoot_Pt
				BMM_Pelvis_ExpTM.exposeNode = bipRoot
				BMM_Ribcage_ExpTM.exposeNode = bNeck

				-- // Hide and freeze helpers, and unhide the visualizer markers
				hide a_BMMobjs
				freeze a_BMMobjs
				unhide BMM_LFoot_Line; unhide BMM_RFoot_Line
				unhide BMM_LowerMass_Line; unhide BMM_UpperMass_Line
				unhide BMM_LFoot_MassLine; unhide BMM_RFoot_MassLine
				unhide BMM_FootCentre_Dot; unhide BMM_MassCentre_Dot
				unhide BMM_UpperToLower_Line; unhide BMM_FootToFoot_Line
				
				--// Move the foot pointers back to zero in Z (as they would be offset due to scaling)
				local BMM_LFoot_Pt = g_LBTools_BMM.f_getNode a_BMMobjs bipRoot "BMM_LFoot_Pt"
				local offsetVal = BMM_LFoot_Pt.transform[4][3]
				move BMM_LFoot_Pt [0,0,-offsetVal]
				
				local BMM_RFoot_Pt = g_LBTools_BMM.f_getNode a_BMMobjs bipRoot "BMM_RFoot_Pt"
				local offsetVal = BMM_RFoot_Pt.transform[4][3]
				move BMM_RFoot_Pt [0,0,-offsetVal]
			)
			else
			(
				delete selection
				messageBox "Error: A Biped Mass Marker already exists for this Biped." title:""
			)
			
			select bipRoot
			
			-- // Update the floor point
			sliderTime = sliderTime + 1
			sliderTime = sliderTime - 1
		)
		else
		(
			messageBox "Error: Select a single Biped part first." title:""
		)
	)
	else
	(
		messageBox "Error: LBTools_BipedMassMarkers maxfile does not exist. Merge failed." title:""
	)
)

) --// end struct

g_LBTools_BMM = s_LBTools_bipedMassMarkers()


/*********************************************************************************************************************************************************************/
--// MACRO CALLS FOR UI
/*********************************************************************************************************************************************************************/

macroscript bipedMassMarkers
	category:"LB Tools"
	buttontext:"{ BMM }"
 	icon:#("LBTools_BipedMassMarkers",1)
	tooltip:"Biped Mass Markers 1.1"
(
	if keyboard.shiftPressed then
	(
		if (querybox "Do you want to delete all mass markers? Otherwise, pick a specific Biped." title:"Biped Mass Markers") then
		(
			g_LBTools_BMM.f_deleteAllMassMarkers()
		)
		else
		(
			messagebox "Pick a Biped (any part) to delete mass markers from..." title:"Biped Mass Markers"
			local pickedObj = pickobject message:"Pick a Biped (any part) to delete mass markers from..." rubberBand:[0,0,0]
			if pickedObj != undefined AND pickedObj != #escape then
			(
				local bipRoot = g_LBTools_BMM.f_findHierarchyRoot pickedObj
				g_LBTools_BMM.f_deleteBMMlayer bipRoot
			)
			else
				messageBox "Pick object aborted." title:"Biped Mass Markers"
		)
	)
	else
	(
		g_LBTools_BMM.f_bipedMassMarkers()
	)
)

/*********************************************************************************************************************************************************************/